import argparse
import os 
import numpy as np
import glob

def get_value(value):
    ''' Handle inf case '''
    if "inf" in value:
        return float(1.7976931348623157e+308)
    else:
        return float(value)

def save_to_array(f_name):
    ''' 
        Retrieve float values string from dumpped file
        NOTE: The activation values should be delimited with space (" ").
    '''
    _values = []
    with open(f_name, "r") as _reader:
        lines = _reader.readlines()
        shape = lines[0].strip("\n")
        _values.append([get_value(v) for line in lines[1:] for v in line.strip("\n").split(" ") if v != ""])
    return shape, np.array(_values).ravel()

def compare_layer(f1_name, f2_name, metric=0, threshold=0.1):
    '''
    Support three metrics to evaluate the similarity of two output tensors,
        0: Cosine Similarity (recommanded way)
        1: Euclidean Distance
        2: Relative Difference
    '''
    shape1, src1 = save_to_array(f1_name) 
    shape2, src2 = save_to_array(f2_name)

    assert(src1.size == src2.size)
    assert(shape1 == shape2)

    similarity = .0
    if metric == 0:
        # Cosine similarity and normalization
        num = np.dot(src1, src2)
        denom = np.linalg.norm(src1) * np.linalg.norm(src2)
        cos = num / denom
        similarity = 0.5 + 0.5 * cos
    elif metric == 1:
        # Euclidean distance and normlization
        euclidean = np.linalg.norm(src1 - src2) 
        similarity = 1.0 / (1.0 + euclidean)
    elif metric == 2:
        # Relative Difference 
        src2_ = np.where(src2 == 0, np.finfo('f').resolution, src2)
        diff = np.absolute(src1 - src2_)
        rd = diff / np.absolute(src2_)
        similarity = float(rd[np.where(rd < threshold)].size) / rd.size  
    else:
        print("ERROR: Unkown metric for similarity analysis!")
        exit(-1)

    return (shape1, similarity)

def run(args):
    # Remove all existing dumpped files
    files = glob.glob(args.directory + "/*")
    for f in files:
        os.remove(f)

    # Create folder to store log files
    if not os.path.exists("./logs"):
        os.makedirs("./logs")

    # Retrieve layer name
    layer_names = []
    with open("layerNames.txt", "r") as f:
        for l in f.readlines():
            layer_names.append(l.strip("\n").split(":")[1])

    # Get the network output activation of FP32 mode as rule
    # TODO: The inference application should be validated in advance, since
    # os.system doesn't handle error case.
    command = args.application + " 2>&1 >/dev/null"
    os.system(command)

    layer_stick_to_float = ""
    if args.layerStickToFloat != "":
        layer_stick_to_float = args.layerStickToFloat + ","

    for layer in range(args.nblayers):

        # Log file for error track or back check
        log_file = "./logs/trt_layer_" + str(layer)+ ".log"

        # Run loss layer in higher precision mode in subsequent round to
        # get rid of interactive accuracy effect.
        if layer_stick_to_float != "":
            command = args.application + \
                      " --int8" + \
                      " --scopedLayerIndex " + str(layer) + \
                      " --layerStickToFloat "+ layer_stick_to_float.rstrip(',') + \
                      " 2>&1 | tee " + log_file + " >/dev/null"
        else:
            command = args.application + \
                      " --int8" + \
                      " --scopedLayerIndex " + str(layer) + \
                      " 2>&1 | tee " + log_file + " >/dev/null"

        os.system(command)

        # Compute the similarity of the network output of pure FP32 mode and
        # mixed FP32-INT8 mode.
        # Expected file name format,
        #       fp32-<output_name>.txt
        #       int8-<output_name>_<scopedLayerIndex>.txt
        f32_fname = "{}/fp32-{}.txt".format(args.directory, args.output)
        i8_fname = "{}/int8-{}_{}.txt".format(args.directory, args.output, layer)

        if os.path.isfile(f32_fname) and os.path.isfile(i8_fname):

            shape,similarity = compare_layer(f32_fname, i8_fname, args.metric)

            if layer == 0:
                print "{:>24}|{:>24}|{:>24}|{:>24}|". \
                    format("ScopedLayerIndex", "LayerName", "OutputShape", "OutputSimilarity")
            print "{:>24}|{:>24}|{:>24}|{:24.4f}|".format(layer, layer_names[layer], shape, similarity)

        else:
            print "\nFailed to locate the dumped activation data!" 
            print "\nPlease check wehther the following command could be running successfully or not,"
            print "\n\t {} \n".format(command)
            exit(-1)

        # Collect all loss layers
        if similarity < args.threshold:
            print "Choose layer [{}] running higher precison mode because it leads the similarity of network output becomes lower than specified threshold [{} vs {}]!".\
                format(layer, similarity, args.threshold)
            layer_stick_to_float += str(layer) + ","

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-a', '--application', type=str,
            default="./sample_mnist",
            help="Inference application with special command support \
            (--scopedLayerIndex and --layerStickToFloat)",
            required=True)
    parser.add_argument('-n', '--nblayers', type=int,
            default=0,
            help="Number of layers to scope",
            required=True)
    parser.add_argument('-o', '--output', type=str,
            default="prob",
            help="Output layer name (of which result will be dumpped \
            for loss analysis)",
            required=True)
    parser.add_argument('-d', '--directory', type=str,
            default="./results/",
            help="Directory that stores the dumpped activation results")
    parser.add_argument('-m', '--metric', type=int,
            default=0,
            help="Metric to evaluate the similarity of output activation \
            result, support three kinds of metrics \
            [0:Cosine Similarity, 1:Euclidean Distance, 2:Relative Difference]")
    parser.add_argument('-t', '--threshold', type=float,
            default=0.99,
            help="Threshold for loss trigger point. If the similarity of \
            current layer is bigger than this threshold, we treat it \
            as a loss layer and choose running this layer in higher precision mode",
            required=True)
    parser.add_argument('-l', '--layerStickToFloat', type=str,
            default="",
            help="Force certain layer running FP32 mode. For multiplier case, \
            use comma as delimiter.")

    args = parser.parse_args()

    run(args)
